"""SweepデータやTODをKID毎に分割されたファイル(*.pkl)にする

Filename: make_divided_data.py
(C) 2021 内藤システムズ

コマンド引数
------------
kids listファイルの名前
local sweepファイルの名前
観測データfitsファイルの名前
一時ファイルを格納するディレクトリ名

使い方
------
コマンドラインでkids.list、Sweepデータ、TODのファイル(fits)名を指定する。
KID毎に分割されたファイル(*.pkl)はcacheディレクトリ内に作成される。

例)
  $ python make_divided_data.py ../raw_dataset/obs/cosmos_20171103184836/kids.list \
                                ../raw_dataset/obs/cosmos_20171103184836/localsweep.sweep \
                                ../raw_dataset/obs/cosmos_20171103184836/20171103184836.fits \
                                cache/20171103184836
"""
import os
import sys
import shutil
import pickle
from . import demerge

def main() -> None:
    """
    コマンドライン引数
    ------------------
    args[1] string kidsリストファイルへのパス(テキストファイル)
    args[2] string ローカルsweepのデータファイルへのパス(テキストファイル)
    args[3] string TODファイルへのパス(fitsファイル)
    args[4] string 分割されたファイルの保管先ディレクトリ
    """
    args = sys.argv
    kidslist = demerge.load_kidslist(args[1])
    localsweeps = demerge.load_localsweep(args[2], framelen=kidslist[0]['framelen'])
    tods = demerge.TODs(args[3])
    kids = demerge.divide_data(kidslist, localsweeps, tods)
    tods.close()
    if os.path.exists(args[4]):
        shutil.rmtree(args[4])
    os.mkdir(args[4])
    for kid in kids:
        with open(args[4] + '/kid{:05}.pkl'.format(int(kid['kidid'])), 'wb') as f:
            pickle.dump(kid, f)

if __name__=='__main__':
    main()



__all__ = [
    "calc_resonance_params",
    "demerge",
    "make_divided_data",
    "make_reduced_fits",
    "merge_function",
    "merge_to_dems",
    "plot",
]
__version__ = "2.5.0"


# submodules
from . import calc_resonance_params
from . import demerge
from . import make_divided_data
from . import make_reduced_fits
from . import merge_function
from . import merge_to_dems
from . import plot


def main() -> None:
    """Run run.sh (works on Linux-like systems)."""
    from pathlib import Path
    from subprocess import run
    from sys import argv

    cmd = Path(__file__).parent / "run.sh"
    run([cmd, *argv[1:]])



"""フィットの結果からphaseなどを計算してfitsに格納する

Filename: make_reduced_fits.py
(C) 2021 内藤システムズ

使い方
------
このスクリプトは第一引数与えられたディレクトリの中のkidXXXXX.pklファイルをすべて読み込む。
各pklファイルにはフィットの結果が格納されていることが正常に動作するための前提条件となる。
解析結果をまとめて第二引数に与えられるファイル名で保存する。

例) カレントディレクトリにあるcacheディレクトリの中身をhoge.fitsへまとめる。
  $ python make_reduced_fits.py ./cache hoge.fits
"""
import os
import sys
import glob
import shutil
import pickle
import numpy as np
from astropy.io import fits
from scipy import interpolate
from . import demerge

def make_reduced_fits(kidfiles, output_filename):
    fitinfo = [] # [[kidid, pread, fc, yfc, linyfc, fr, dfr, Qr, dQr, Qc, dQc, Qi, dQi], ...]
    readout = demerge.readout_dict()
    kidsinfo = demerge.kids_dict()
    readout['hdr_val_lis'][1] = 'fuga'
    kidsinfo['hdr_val_lis'][1] = 'localsweep.sweep'
    downsampling_rate = 1

    framert = None
    npoints = None
    for kidfile in kidfiles:
        with open(kidfile, 'rb') as f:
            kid = pickle.load(f)
        timestamp = kid['tod_timestamp'][:-100] #データの最後の部分にはよくないデータが含まれる可能性があるので、最後の100個ぐらいを除外する。
        ts = demerge.rebin_array(timestamp, downsampling_rate) #1はdown samplingしないことを意味する
        fc = kid['tod'].frequency

        if kid['resonance_params'] is None:
            fr = np.nan
            dfr = np.nan
            Qr = np.nan
            dQr = np.nan
            Qc = np.nan
            dQc = np.nan
            Qi = np.nan
            dQi = np.nan
            yfc = np.nan
            linyfc = np.nan
            ampl = np.array([np.nan for j in range(len(ts))])
            phase = np.array([np.nan for j in range(len(ts))])
            linphase = np.array([np.nan for j in range(len(ts))])
            rewound_tod = None
        else:
            fr = kid['resonance_params']['fr'].value # GHz
            dfr = kid['resonance_params']['fr'].stderr # GHz
            Qr = kid['resonance_params']['Qr'].value
            dQr = kid['resonance_params']['Qr'].stderr
            Qc = kid['resonance_params']['Qc'].value
            dQc = kid['resonance_params']['Qc'].stderr
            Qi = kid['resonance_params']['Qi'].value
            dQi = kid['resonance_params']['Qi'].stderr

            arga = kid['resonance_params']['arga']
            absa = kid['resonance_params']['absa']
            tau = kid['resonance_params']['tau']
            phi0 = kid['resonance_params']['phi0']
            c = kid['resonance_params']['c']

            rewound_localsweep = demerge.gaolinbg_rewind(kid['localsweep'].x, kid['localsweep'].iq, arga, absa, tau, fr, Qr, Qc, phi0, c)
            fitted_localsweep = demerge.gaolinbg_function()(kid['localsweep'].x, arga, absa, tau, fr, Qr, Qc, phi0, c)
            rewound_fitted_localsweep = demerge.gaolinbg_rewind(kid['localsweep'].x, fitted_localsweep, arga, absa, tau, fr, Qr, Qc, phi0, c)

            phase_localsweep = -np.angle(-rewound_localsweep)
            phase_fitted_localsweep = -np.angle(-rewound_fitted_localsweep)

            ## spline interpolation
            tck = interpolate.splrep(kid['localsweep'].x, phase_localsweep, s=0)
            yfc = interpolate.splev(fc, tck, der=0) # phase of carrier f
            #yfr = interpolate.splev(fr, tck, der=0) # phase of resonance f

            ## linphaseの計算(linyfc = k.convert_to_fshift(yfc, opt='linphase'))
            tck = interpolate.splrep(phase_fitted_localsweep, kid['localsweep'].x, s=0)
            f = interpolate.splev(yfc, tck, der=0)
            linyfc = 4.0*Qr*(f - fr)/fr

            # 抜粋:mkid_data.data.FixedData::calibrate_with_blind_tones()メソッド
            GHz = 1e9
            calibrated = demerge.calibrate_with_blind_tones(kid['tod'], kid['blind_tone_left'], kid['blind_tone_right'])
            #calibrated_tod = FixedData(kid['tod'].frequency*GHz, kid['tod'].t, np.real(calibrated), np.imag(calibrated), info=kid['tod'].info)
            # calibrated_tod = demerge.FixedData()
            # calibrated_tod.timestamp = kid['tod'].t
            # calibrated_tod.i = np.real(calibrated)
            # calibrated_tod.q = np.imag(calibrated)
            # calibrated_tod._frequency = kid['tod'].frequency*GHz
            # calibrated_tod.info = kid['tod'].info

            # 抜粋:mkid_data.kidfit.KidFitResult::rewound_ampl_phase()メソッド
            rw = demerge.gaolinbg_rewind(kid['tod'].frequency, calibrated, arga, absa, tau, fr, Qr, Qc, phi0, c)
            #rw = demerge.gaolinbg_rewind(calibrated_tod.frequency, calibrated_tod.iq, arga, absa, tau, fr, Qr, Qc, phi0, c)
            ampl  = 2*np.abs(rw)
            phase = -np.angle(-rw)
            index_ar0 = np.where(abs(phase)>3.1)
            if len(index_ar0[0])>0:
                index0 = index_ar0[0][0]
                initial = phase[index0]
                if initial>0.:
                    index_ar1 = np.where(phase<0.)
                    factor = 2.*np.pi
                else:
                    index_ar1 = np.where(phase>0.)
                    factor = -2.*np.pi
                for index in index_ar1[0]:
                    if index>index0: phase[index] += factor

            #rewound_tod = FixedFitData(calibrated_tod.frequency, calibrated_tod.t, ampl, phase, info=calibrated_tod.info)
            #rewound_tod = FixedFitData(kid['tod'].frequency, kid['tod'].t, ampl, phase, info=kid['tod'].info)
            rewound_tod = demerge.FixedFitData(kid['tod'].frequency, kid['tod'].t, ampl, phase, info=kid['tod'].info)

            #ts, ampl, phase = k.get_cache('deglitch').unpack()
            baseline_thresh = 6.0
            glitch_thresh = 5.0
            clusterize_thresh = 2
            interp_offset = 0
            bad = demerge.find_glitch([rewound_tod.phase], baseline_thresh, glitch_thresh, clusterize_thresh, interp_offset)

            # 抜粋:bbsweeplib.kids.KID::deglitch()
            ampl = demerge.interpolate_bad(rewound_tod.amplitude, bad)
            phase = demerge.interpolate_bad(rewound_tod.phase, bad)
            d = demerge.FixedFitData(rewound_tod.frequency*GHz, rewound_tod.t, ampl, phase)
            ts, ampl, phase = d.unpack()

            #tck = interpolate.splrep(phase_fitted_localsweep, kid['localsweep'].x, s=0)
            f = interpolate.splev(phase, tck, der=0)
            linphase = 4.0*Qr*(f - fr)/fr
            ts = demerge.rebin_array(ts, downsampling_rate)
            ampl = demerge.rebin_array(ampl, downsampling_rate)
            phase = demerge.rebin_array(phase, downsampling_rate)

        fitinfo.append([kid['kidid'], kid['readpower'], fc, yfc, linyfc, fr, dfr, Qr, dQr, Qc, dQc, Qi, dQi])

        if len(readout['cols_data_lis']) == 0:
            readout['cols_data_lis'].append(timestamp)
            readout['cols_data_lis'].append(np.ones(len(timestamp))*kid['pixelid'])

            framert = kid['framert']
            npoints = kid['npoints']

        readout['cols_key_lis'].append('Amp, Ph, linPh {}'.format(kid['kidid']))
        readout['cols_data_lis'].append(np.array([ampl, phase, linphase]).T)
        readout['tform'].append('3E')
        readout['tunit'].append(None)

        # plot.pyのtodのプロットで使うために保存
        kid['amp'] = ampl
        kid['ph'] = phase
        kid['linph'] = linphase
        kid['yfc'] = yfc
        kid['linyfc'] = linyfc
        kid['rewound_tod'] = rewound_tod
        with open(kidfile, 'wb') as f:
            pickle.dump(kid, f)

        readout['hdr_com_lis'].append('label for field {}'.format(kid['kidid'] + 2))
        readout['hdr_com_lis'].append('data format of field {}'.format(kid['kidid'] + 2))
        continue

    pixelid = 0
    #nkid = len(glob.glob('cache/kid*.pkl'))
    nkid = len(kidfiles)
    readout['hdr_val_lis'][2] = framert
    readout['hdr_val_lis'][3] = npoints
    readout['hdr_val_lis'][4] = downsampling_rate
    readout['hdr_key_lis'].append('NKID{}'.format(pixelid))
    readout['hdr_val_lis'].append(nkid) #全KID数
    readout['hdr_com_lis'].insert(5, 'number of KIDs (pixel {})'.format(pixelid))

    kidsinfo['hdr_key_lis'].append('NKID{}'.format(pixelid))
    kidsinfo['hdr_val_lis'].append(nkid)
    kidsinfo['hdr_com_lis'].insert(2, 'number of KIDs (pixel {})'.format(pixelid))

    kidid_, pread_, fc_, yfc_, linyfc_, fr_, dfr_, Qr_, dQr_, Qc_, dQc_, Qi_, dQi_ = [np.array(fitinfo)[:,i] for i in range(13)]
    kidsinfo['cols_data_lis'].append(np.ones(nkid)*pixelid)
    kidsinfo['cols_data_lis'].append(kidid_)
    kidsinfo['cols_data_lis'].append(pread_)
    kidsinfo['cols_data_lis'].append(fc_)
    kidsinfo['cols_data_lis'].append(np.array([yfc_, linyfc_]).T)
    kidsinfo['cols_data_lis'].append(np.array([fr_, dfr_]).T)
    kidsinfo['cols_data_lis'].append(np.array([Qr_, dQr_]).T)
    kidsinfo['cols_data_lis'].append(np.array([Qc_, dQc_]).T)
    kidsinfo['cols_data_lis'].append(np.array([Qi_, dQi_]).T)

    hdus = fits.HDUList()
    hdus.append(fits.PrimaryHDU())
    hdus.append(demerge.createBinTableHDU(kidsinfo))
    hdus.writeto(output_filename)
    hdus.close()

    readout_hdu = demerge.createBinTableHDU(readout)
    fits.append(output_filename, readout_hdu.data, readout_hdu.header, memmap=True)
    return

def main() -> None:
    """
    コマンドライン引数
    ------------------
    args[1] string pklファイルが格納されているcacheディレクトリへのパス
    args[2] string 出力ファイル名
    """
    args = sys.argv
    cache_dir = args[1]
    output_filename = args[2]
    kidfiles = sorted(glob.glob(cache_dir + '/kid*.pkl')) #ファイル名で昇順で並べ替える
    make_reduced_fits(kidfiles, output_filename)

if __name__=='__main__':
    main()



"""merge_to_dfits.py: Read logging data and merge them into a FITS object

 Author : Tetsutaro Ueda, Junya Suzuki, Kenichi Karatsu, Tatsuya Takekoshi
 Created: 2017/11/02
 Revision History:
     2018/02/08 - KK - rewrite using class.
     2018/06/08 - TT - apply new calibration method.
     2021         NAITO systems modfied.
     2023         NAITO systems modfied.
"""
from __future__ import print_function

__all__ = [
    'FORM_FITSTIME',
    'FORM_FITSTIME_P',
    'DEFAULT_ROOM_T',
    'create_bintablehdu',
    'load_obsinst',
    'get_maskid_corresp'
    'Tlos_model',
    'calibrate_to_power',
    'convert_asciitime',
    'convert_timestamp'
]

from datetime import datetime
from datetime import timedelta
from calendar import timegm
from astropy.io import fits, ascii
import numpy as np
import scipy.interpolate
import sys

#-------------------------------- CONSTANTS
FORM_FITSTIME   = '%Y-%m-%dT%H:%M:%S'                          # YYYY-mm-ddTHH:MM:SS
FORM_FITSTIME_P = '%Y-%m-%dT%H:%M:%S.%f'                       # YYYY-mm-ddTHH:MM:SS.ss

CABIN_Q_MARGIN  = 5*60 # seconds. Margin for cabin data query.
DEFAULT_ROOM_T  = 17. + 273. # Kelvin
DEFAULT_AMB_T   = 0.  + 273. # Kelvin

def create_bintablehdu(hd):
    """Create Binary Table HDU from 'hdu_dict'"""
    header = fits.Header()
    for (i, j) in zip(hd['hdr_vals'].items(), hd['hdr_coms'].items()):
        header[i[0]] = i[1], j[1]
    columns = [
        fits.Column(name=i[0], format=j[1], array=i[1], unit=k[1])
        for (i, j, k) in zip(
            hd['col_vals'].items(),
            hd['col_form'].items(),
            hd['col_unit'].items()
        )
    ]
    hdu = fits.BinTableHDU.from_columns(columns, header)
    for i in hd['hdr_coms'].items():
        hdu.header.comments[i[0]] = i[1]
    return hdu

def load_obsinst(obsinst):
    """Get data for 'OBSINFO'"""
    if not '.obs' in obsinst:
        raise ValueError('The input file must be an observational instruction!!')

    with open(obsinst, 'r') as f:
        equinox = 2000  # Default parameter
        for line in f:
            if 'SET ANTENNA_G TRK_TYPE' in line:
                trktype = line.split()[-1].strip('\'')
            elif 'SET ANTENNA_G SRC_NAME' in line:
                obs_object = line.split()[-1].strip('\'')
            elif 'SET ANTENNA_G SRC_POS' in line:
                srcpos = [float(c) for c in line.split()[-1].strip('()').split(',')]
            elif 'SET ANTENNA_G EPOCH' in line:
                equinox = line.split()[-1].strip('\'JB')
            elif 'SET DES OBS_USER' in line:
                observer = line.split()[-1].strip('\'')
            elif 'SET DES PROJECT' in line:
                project = line.split()[-1].strip('\'')
            elif 'SET DES PROJECT' in line:
                project = line.split()[-1].strip('\'')
            elif '% OBS=' in line:
                observation = line.split('=')[-1].strip()
    if trktype == 'RADEC':
        ra  = srcpos[0]
        dec = srcpos[1]
    else:
        ra  = 0
        dec = 0
    return {'observer': observer, 'obs_object': obs_object,  'ra': ra, 'dec': dec, 'equinox': equinox, 'project': project, 'observation': observation}

def get_maskid_corresp(pixelid, ddb):
    """Get Correspondance of 'master' and 'kid'"""
    nkid = ddb['KIDFILT'].header['NKID%d' %pixelid]
    kiddict, kidfilt = {}, {}
    for (i, j, k, l) in zip(ddb['KIDFILT'].data['kidid'],
                            ddb['KIDFILT'].data['masterid'],
                            ddb['KIDFILT'].data['F_filter, dF_filter'],
                            ddb['KIDFILT'].data['Q_filter, dQ_filter']):
        kiddict[i] = j
        kidfilt[i] = (k[0], l[0])
    kidname = {}
    for (i, j) in zip(ddb['KIDDES'].data['masterid'], ddb['KIDDES'].data['attribute']):
        kidname[i] = j

    masterids, kidids, kidtypes, kidfreqs, kidQs = [], [], [], [], []
    for i in range(nkid):
        masterid = kiddict[i]
        if masterid < 0:
            kind = 'unknown'
        else:
            kind = kidname[masterid]

        if kind=='wideband': attr = 0
        elif kind=='filter': attr = 1
        elif kind=='blind':  attr = 2
        elif kind=='Al':     attr = 3
        elif kind=='NbTiN':  attr = 4
        else:                attr = -1

        masterids.append( masterid )
        kidids.append( i )
        kidtypes.append( attr )
        kidfreqs.append( kidfilt[i][0] )
        kidQs.append( kidfilt[i][1] )
    return masterids, kidids, kidtypes, kidfreqs, kidQs

def Tlos_model(dx, p0, etaf, T0, Troom, Tamb):
    """Calibrate 'amplitude' and 'phase' to 'power'"""
    return (dx + p0*np.sqrt(Troom+T0))**2 / (p0**2 * etaf) - T0/etaf - (1-etaf)/etaf*Tamb

def calibrate_to_power(pixelid, Troom, Tamb, rhdus, ddb):
    nkid = rhdus['READOUT'].header['NKID%d' %pixelid]
    kiddict = {}
    for (i, j) in zip(ddb['KIDFILT'].data['kidid'], ddb['KIDFILT'].data['masterid']):
        kiddict[i] = j

    linphase = np.transpose([rhdus['READOUT'].data['Amp, Ph, linPh %d' %i].T[2] for i in range(nkid)])
    linyfc   = rhdus['KIDSINFO'].data['yfc, linyfc'].T[1]
    Qr       = rhdus['KIDSINFO'].data['Qr, dQr (300K)'].T[0]
    fshift = np.array((linphase - linyfc)/ (4.*Qr)).T
    fshift_err = np.zeros( len(fshift) )
    #---- Responsivity curve
    (p0, etaf, T0) = ddb['KIDRESP'].data['cal params'].T
    Tsignal = []
    for i in range(nkid):
        masterid = kiddict[i]
        if masterid<0:
            Tsignal.append( [np.nan for j in range( len(fshift[i]) )] )
            continue
        #---- Convert to power
        Tsignal.append(Tlos_model(fshift[i], p0[i], etaf[i], T0[i], Troom, Tamb))
    return np.array(Tsignal).T

def convert_asciitime(asciitime, form_fitstime):
    """Ascii time"""
    asciitime = [datetime.strptime('%14.6f' %t, '%Y%m%d%H%M%S.%f') for t in asciitime]
    asciitime = [datetime.strftime(t, form_fitstime) for t in asciitime]
    return np.array(asciitime)

def convert_timestamp(timestamp):
    """Timestamp"""
    timestamp = [datetime.utcfromtimestamp(t) for t in timestamp]
    timestamp = [datetime.strftime(t, FORM_FITSTIME_P) for t in timestamp]
    return np.array(timestamp)

def retrieve_cabin_temps(filename):
    """キャビン内温度を取得する
    引数
    ====
    str ファイル名

    戻り値
    ======
    tuple (timestames, upperCabinTemps, lowerCabinTemps)
      tupleの各要素はnumpy.array。要素数は同じ。
    """
    table = ascii.read(filename, format='no_header')

    # 日付と時刻を取得して文字列でタイムスタンプを作成しそれをnumpy.datetime64へ変換する
    # テーブルの1列目と2列目がそれぞれ日付と時刻
    datetimes = []
    for date, time in zip(table['col1'], table['col2']):
        s = '{}T{}'.format(date, time)
        s = s.replace('/', '-')
        datetimes.append(s)
    datetimes       = np.array(datetimes).astype('datetime64[ns]')
    upper_cabin_temps = np.array(table['col3']).astype(np.float64)
    lower_cabin_temps = np.array(table['col4']).astype(np.float64)

    return (datetimes, upper_cabin_temps, lower_cabin_temps)

def retrieve_skychop_states(filename):
    """skychopファイルからskychopの時系列状態を取得する
    引数
    ====
    str ファイル名

    戻り値
    ======
    tuple (timestames, states)
      tupleの各要素はnumpy.array。要素数は同じ。

    時刻について
    ============
    skychopファイルに記録されている時刻はUNIX時間。

    ファイル形式
    ============
    1列目 UNIX時刻
    2列目 0/1による状態
    "#"から始まるコメントがファイル冒頭に数行ある。
    """
    table = ascii.read(filename, guess=False, format='no_header', delimiter=' ', names=['datetime', 'state'])

    datetimes = np.array(table['datetime']).astype(np.float64)
    states    = np.array(table['state']).astype(np.int8)
    return (datetimes, states)

def retrieve_misti_log(filename):
    """mistiファイルからの時系列データを取得する
    引数
    ====
    str ファイル名

    戻り値
    ======
    tuple (timestames, az, el, pwv)
      tupleの各要素はnumpy.array。要素数は同じ。

    ファイル形式
    ============
    1列目 年/月/日
    2列目 時:分:6列目 秒(小数点以下2桁も含む)
    3列目 az(deg)
    4列目 el(deg)
    5列目 pwv(um)
    6列目 Tground(K)
    "#"から始まるコメントがファイル冒頭に数行ある。
    """
    column_names = [
        'date',
        'time',
        'az',
        'el',
        'pwv',
        'Tround',
    ]
    table = ascii.read(filename, guess=False, format='no_header', delimiter=' ', names=column_names)

    az  = np.array(table['az']).astype(np.float64)
    el  = np.array(table['el']).astype(np.float64)
    pwv = np.array(table['pwv']).astype(np.float64)/1000.0 # umからmmへ変換

    datetimes = []
    for row in table:
        datetimes.append(datetime.strptime('{} {}'.format(row['date'], row['time']), '%Y/%m/%d %H:%M:%S.%f'))

    return (np.array(datetimes).astype('datetime64[ns]'), az, el, pwv)



"""demsオブジェクトを生成する

python 3.9
dems   0.4.0

(C) 2023 内藤システムズ
"""
import argparse
import numpy          as np
import xarray         as xr
from . import merge_function as mf

from astropy.io import fits, ascii
from dems.d2    import MS


# constants
from . import __version__ as DEMERGE_VERSION


def merge_to_dems(
        ddbfits_path='',
        obsinst_path='',
        antenna_path='',
        readout_path='',
        skychop_path='',
        weather_path='',
        misti_path='',
        cabin_path='',
        **kwargs
        ):
    # その他の引数の処理と既定値の設定
    pixel_id   = kwargs.pop('pixel_id',   0)
    coordinate = kwargs.pop('coordinate', 'azel')
    mode       = kwargs.pop('mode',   0)
    loadtype   = kwargs.pop('loadtype',   'Tsignal')
    # find R, sky
    findR  = kwargs.pop("findR",  False)
    ch     = kwargs.pop("ch",     0)
    Rth    = kwargs.pop("Rth",    280)
    skyth  = kwargs.pop("skyth",  150)
    cutnum = kwargs.pop("cutnum", 1)
    # still
    still  = kwargs.pop("still",  False)
    period = kwargs.pop("period", 2) # 秒
    # shuttle
    shuttle     = kwargs.pop("shuttle", False)
    lon_min_off = kwargs.pop("lon_min_off", 0)
    lon_max_off = kwargs.pop("lon_max_off", 0)
    lon_min_on  = kwargs.pop("lon_min_on",  0)
    lon_max_on  = kwargs.pop("lon_max_on",  0)

    # 時刻と各種データを読み込む(必要に応じて時刻はnp.datetime64[ns]へ変換する)
    readout_hdul   = fits.open(readout_path)
    ddbfits_hdul   = fits.open(ddbfits_path)
    weather_table  = ascii.read(weather_path)
    antenna_table  = ascii.read(antenna_path)[:-1] # 最後の1行は終端を表す意味のないデータが入っているため無視する
    obsinst_params = mf.load_obsinst(obsinst_path) # 観測スクリプトに含まれているパラメタを抽出する

    times = mf.convert_timestamp(readout_hdul['READOUT'].data['timestamp'])
    times = np.array(times).astype('datetime64[ns]')

    times_misti, az_misti, el_misti, pwv_misti = mf.retrieve_misti_log(misti_path)

    times_cabin, upper_cabin_temp, lower_cabin_temp = mf.retrieve_cabin_temps(cabin_path)
    lower_cabin_temp = lower_cabin_temp + 273.15 # 度CからKへ変換

    times_weather = mf.convert_asciitime(weather_table['time'], '%Y-%m-%dT%H:%M:%S.%f')
    times_weather = np.array(times_weather).astype('datetime64[ns]')

    times_skychop, states_skychop = mf.retrieve_skychop_states(skychop_path)
    times_skychop = mf.convert_timestamp(times_skychop)
    times_skychop = np.array(times_skychop).astype('datetime64[ns]')

    times_antenna = mf.convert_asciitime(antenna_table['time'], '%Y-%m-%dT%H:%M:%S.%f')
    times_antenna = np.array(times_antenna).astype('datetime64[ns]')

    response = None
    if loadtype == 'Tsignal':
        # T_signalsを計算する
        master_id, kid_id, kid_type, kid_freq, kid_Q = mf.get_maskid_corresp(pixel_id, ddbfits_hdul)
        T_amb     = np.nanmean(weather_table['tmperature']) + 273.15 # 度CからKへ変換
        T_signals = mf.calibrate_to_power(pixel_id, lower_cabin_temp[0], T_amb, readout_hdul, ddbfits_hdul)
        response  = T_signals
    else:
        raise KeyError('Invalid loadtype: {}'.format(loadtype))

    ddbfits_hdul.close()
    readout_hdul.close()

    # モードに応じて経度(lon)と緯度(lat)を選択(azelかradecか)する
    if coordinate == 'azel':
        if   'az-prg(no-cor)' in antenna_table.colnames:
            az_prg  = antenna_table['az-prg(no-cor)']
            el_prog = antenna_table['el-prog(no-cor)']
        elif 'az-prg(no-col)' in antenna_table.colnames:
            az_prg  = antenna_table['az-prg(no-col)']
            el_prog = antenna_table['el-prog(no-col)']
        else:
            raise KeyError('{}ファイルにaz-prg(no-cor)列とaz-prg(no-col)列がありません。どちらか片方が存在する必要があります。ANTENNAファイルの形式を確認してください。'.format(antenna_path))

        lon        = az_prg  + antenna_table['az-real'] - antenna_table['az-prg']
        lat        = el_prog + antenna_table['el-real'] - antenna_table['el-prg']
        lon_origin = np.median(antenna_table['az-prog(center)'])
        lat_origin = np.median(antenna_table['el-prog(center)'])
        if mode in [0, 1]:
            lon -= antenna_table['az-prog(center)']
            lat -= antenna_table['el-prog(center)']
            if mode == 0:
                lon *= np.cos(np.deg2rad(lat))
    elif coordinate == 'radec':
        lon        = antenna_table['ra-prg']
        lat        = antenna_table['dec-prg']
        lon_origin = obsinst_params['ra'] # 観測スクリプトに設定されているRA,DEC
        lat_origin = obsinst_params['dec']
    else:
        raise KeyError('Invalid coodinate type: {}'.format(coordinate))

    # 補間関数で扱うためにSCANTYPE(文字列)を適当な整数に対応させる
    states = np.array(antenna_table['type'])
    state_types = {state_type:i for i, state_type in enumerate(np.unique(states))}
    state_type_numbers = np.zeros(states.shape[0], dtype=int)
    for state_type, i in state_types.items():
        state_type_numbers[states == state_type] = i

    # 補間のためにDataArrayへ格納する
    response_xr               = xr.DataArray(data=response, dims=['time', 'chan'], coords=[times, kid_id])
    lon_xr                    = xr.DataArray(data=lon,                             coords={'time': times_antenna})
    lat_xr                    = xr.DataArray(data=lat,                             coords={'time': times_antenna})
    temperature_xr            = xr.DataArray(data=weather_table['tmperature'],     coords={'time': times_weather})
    humidity_xr               = xr.DataArray(data=weather_table['vapor-pressure'], coords={'time': times_weather})
    pressure_xr               = xr.DataArray(data=weather_table['presure'],        coords={'time': times_weather})
    wind_speed_xr             = xr.DataArray(data=weather_table['aux1'],           coords={'time': times_weather})
    wind_direction_xr         = xr.DataArray(data=weather_table['aux2'],           coords={'time': times_weather})
    skychop_state_xr          = xr.DataArray(data=states_skychop,                  coords={'time': times_skychop})
    aste_cabin_temperature_xr = xr.DataArray(data=lower_cabin_temp,                coords={'time': times_cabin})
    aste_misti_lon_xr         = xr.DataArray(data=az_misti,                        coords={'time': times_misti})
    aste_misti_lat_xr         = xr.DataArray(data=el_misti,                        coords={'time': times_misti})
    aste_misti_pwv_xr         = xr.DataArray(data=pwv_misti,                       coords={'time': times_misti})
    state_type_numbers_xr     = xr.DataArray(data=state_type_numbers,              coords={'time': times_antenna})

    # Tsignalsの時刻に合わせて補間する
    lon                    =                    lon_xr.interp_like(response_xr)
    lat                    =                    lat_xr.interp_like(response_xr)
    temperature            =            temperature_xr.interp_like(response_xr)
    humidity               =               humidity_xr.interp_like(response_xr)
    pressure               =               pressure_xr.interp_like(response_xr)
    wind_speed             =             wind_speed_xr.interp_like(response_xr)
    wind_direction         =         wind_direction_xr.interp_like(response_xr)
    aste_cabin_temperature = aste_cabin_temperature_xr.interp_like(response_xr)
    aste_misti_lon         =         aste_misti_lon_xr.interp_like(response_xr)
    aste_misti_lat         =         aste_misti_lat_xr.interp_like(response_xr)
    aste_misti_pwv         =         aste_misti_pwv_xr.interp_like(response_xr)
    skychop_state          =          skychop_state_xr.interp_like(response_xr, method='nearest')
    state_type_numbers     =     state_type_numbers_xr.interp_like(response_xr, method='nearest')

    # 補間後のSTATETYPEを文字列に戻す
    state = np.full_like(state_type_numbers, 'GRAD', dtype='<U8')
    for state_type, i in state_types.items():
        state[state_type_numbers == i] = state_type

    # 静止データの周期に応じてOFFマスクとSCANマスクを設定する
    if still:
        seconds = (times - times[0])/np.timedelta64(1, 's')
        for i in range(int(seconds[-1]) // period + 1):
            off_mask = (period*(2*i)     <= seconds) & (seconds < period*(2*i + 1))
            on_mask  = (period*(2*i + 1) <= seconds) & (seconds < period*(2*i + 2))
            state[off_mask] = 'OFF'
            state[on_mask]  = 'SCAN'

    # shuttle観測のマスクを設定する
    if shuttle:
        mask_off = (lon_min_off < lon) & (lon < lon_max_off)
        mask_on  = (lon_min_on  < lon) & (lon < lon_max_on)
        state[mask_off]                 = 'OFF'
        state[mask_on]                  = 'SCAN'
        state[(~mask_off) & (~mask_on)] = 'JUNK'

    # Rとskyの部分を探し、その変化点も含めてJUNKな部分を調べる。
    if findR:
        # Rの部分とその変化の部分を探す
        indices = np.where(response[:, ch] >= Rth)
        state[indices] = 'R'

        # cutnum個だけ左右を切り取った配列を作り、互いに異なる部分を探す。そこはおおよそ変化が起きている部分と考えられる。
        #
        # cutnum = 2 の例
        #
        # ** 状態Rへ変化していく場合 **
        # state                   = 0 0 0 0 0 1 1 1 1 1
        # [cutnum: ]              = 0 0 0 1 1 1 1 1
        # [:-cutnum]              = 0 0 0 0 0 1 1 1
        # [cutnum:] != [:-cutnum] = 0 0 0 1 1 0 0 0
        # state_right_shift       = 0 0 0 0 0 1 1 0 0 0
        # state_left_shift        = 0 0 0 1 1 0 0 0 0 0
        # state_R                 = 0 0 0 0 0 1 1 1 1 1
        # mask_moving             = 0 0 0 0 0 1 1 0 0 0
        #
        # ** 状態Rから別の状態へ変化していく場合 **
        # state                   = 1 1 1 1 1 0 0 0 0 0
        # [cutnum: ]              = 1 1 1 0 0 0 0 0
        # [:-cutnum]              = 1 1 1 1 1 0 0 0
        # [cutnum:] != [:-cutnum] = 0 0 0 1 1 0 0 0
        # state_right_shift       = 0 0 0 0 0 1 1 0 0 0
        # state_left_shift        = 0 0 0 1 1 0 0 0 0 0
        # state_R                 = 1 1 1 1 1 0 0 0 0 0
        # mask_moving             = 0 0 0 1 1 1 1 0 0 0
        #
        # 状態がRへ変化する場合と、状態Rから別の状態へ変化する場合でmask_movingのでき方が違う。
        #
        state_cut = state[cutnum:] != state[:-cutnum]

        state_right_shift = np.hstack( [[False]*cutnum, state_cut] ) # 左側をFalseで埋めて右にずらす
        state_left_shift  = np.hstack( [state_cut, [False]*cutnum] ) # 右側をFalseで埋めて左にずらす
        state_R           =          ( state == 'R' )

        mask_moving = state_R & state_left_shift | state_right_shift
        state[mask_moving] = 'JUNK'

        indices = (response[:, ch] >  skyth) & (state != 'R')
        state[indices] = 'JUNK'

        indices = (response[:, ch] <= skyth) & (state == 'R')
        state[indices] = 'JUNK'

        # SKYの部分とその変化の部分を探す
        indices      = np.where(response[:, ch] <= skyth)
        tmp          = state.copy() # 最終的にSKYを残さないためにコピーを扱う
        tmp[indices] = 'SKY'        # 一時的にSKYをマークする

        tmp_cut = tmp[cutnum:] != tmp[:-cutnum] # cutnum個だけ左右にずらした配列を作り、変化を探す。

        tmp_right_shift = np.hstack( [[False]*cutnum, tmp_cut] )
        tmp_left_shift  = np.hstack( [tmp_cut, [False]*cutnum] )
        tmp_sky         =          ( tmp == 'SKY' )

        mask_moving = tmp_sky & tmp_left_shift | tmp_right_shift
        state[mask_moving] = 'JUNK' # 変化の部分はJUNKに置き換える(Rとは違いSKYは残らない)

    return MS.new(
        data                    =response,
        time                    =times,
        chan                    =kid_id,
        state                   =state,
        lon                     =lon,
        lat                     =lat,
        lon_origin              =lon_origin,
        lat_origin              =lat_origin,
        temperature             =temperature,
        pressure                =pressure,
        humidity                =humidity,
        wind_speed              =wind_speed,
        wind_direction          =wind_direction,
        aste_cabin_temperature  =aste_cabin_temperature,
        aste_misti_lon          =aste_misti_lon,
        aste_misti_lat          =aste_misti_lat,
        aste_misti_pwv          =aste_misti_pwv,
        d2_mkid_id              =kid_id,
        d2_mkid_type            =kid_type,
        d2_mkid_frequency       =kid_freq,
        d2_skychopper_isblocking=skychop_state,
        d2_demerge_version       =DEMERGE_VERSION,
        beam_major              =0.005,  # 18 arcsec MergeToDfits()でも固定値が指定されていた
        beam_minor              =0.005,  # 18 arcsec MergeToDfits()でも固定値が指定されていた
        beam_pa                 =0.005,  # 18 arcsec MergeToDfits()でも固定値が指定されていた
        exposure                =1./196, #           MergeToDfits()でも固定値が指定されていた
        interval                =1./196, #           MergeToDfits()でも固定値が指定されていた
        observation             =obsinst_params['observation'],
        observer                =obsinst_params['observer'],
        object                  =obsinst_params['obs_object'],
    )

def main() -> None:
    """Demsオブジェクトを作成する"""
    parser = argparse.ArgumentParser()

    # 必須引数
    parser.add_argument('filename',  type=str, help='出力ファイルへのパスを指定して下さい(.zarr.zip)')
    parser.add_argument('--ddb',     type=str, required=True, help='DDBファイルへのパスを指定して下さい(.fits.gz)')
    parser.add_argument('--obs',     type=str, required=True, help='obsファイルへのパスを指定して下さい(.obs)')
    parser.add_argument('--antenna', type=str, required=True, help='antennaファイルへのパスを指定して下さい(.antenna)')
    parser.add_argument('--readout', type=str, required=True, help='reduced readoutファイルへのパスを指定して下さい(.fits)')
    parser.add_argument('--skychop', type=str, required=True, help='skychopファイルへのパスを指定して下さい(.skychop)')
    parser.add_argument('--weather', type=str, required=True, help='weatherファイルへのパスを指定して下さい(.weather)')
    parser.add_argument('--misti',   type=str, required=True, help='mistiファイルへのパスを指定して下さい(.misti)')
    parser.add_argument('--cabin',   type=str, required=True, help='cabinファイルへのパスを指定して下さい(.cabin)')

    # オプション引数
    parser.add_argument('--pixel_id',    type=int,   default=0,         help='pixel_idを整数で指定します')
    parser.add_argument('--coordinate',  type=str,   default='azel',    help='座標系(azel/radec)を文字列で指定します')
    parser.add_argument('--mode',        type=int,   default=0,         help='座標の読み込み方法を整数で指定します(0:相対座標cos射影あり(おすすめ既定値), 1:相対座標cos射影なし, 2:絶対座標)')
    parser.add_argument('--loadtype',    type=str,   default='Tsignal', help='読み込むデータを文字列で指定します(既定値: Tsignal, 現在はTsignalの読み込みしかできません)')
    parser.add_argument('--findR',       type=bool,  default=False,     help='findRにTrueを指定するとFindR, Skyを実行します')
    parser.add_argument('--ch',          type=int,   default=0,         help='findRに利用するチャネルを整数で指定します')
    parser.add_argument('--Rth',         type=float, default=280.0,     help='R閾値を実数で指定します')
    parser.add_argument('--skyth',       type=float, default=150.0,     help='sky閾値を実数で指定します')
    parser.add_argument('--cutnum',      type=int,   default=1,         help='findRでのカット数を整数で指定します')
    parser.add_argument('--still',       type=bool,  default=False,     help='Trueに設定するとstill観測用の解析を行います')
    parser.add_argument('--period',      type=int,   default=2,         help='still観測の1/2周期(秒)を整数で指定します')
    parser.add_argument('--shuttle',     type=bool,  default=False,     help='Trueを指定するとshuttle観測用の解析を行います')
    parser.add_argument('--lon_min_off', type=float, default=0.0,       help='shuttle観測時のOFFにするlongitudeの最小値を実数で指定します')
    parser.add_argument('--lon_max_off', type=float, default=0.0,       help='shuttle観測時のOFFにするlongitudeの最大値を実数で指定します')
    parser.add_argument('--lon_min_on',  type=float, default=0.0,       help='shuttle観測時のONにするlongitudeの最小値を実数で指定します')
    parser.add_argument('--lon_max_on',  type=float, default=0.0,       help='shuttle観測時のONにするlongitudeの最大値を実数で指定します')

    a = parser.parse_args()

    dems = merge_to_dems(
        ddbfits_path=a.ddb,
        obsinst_path=a.obs,
        antenna_path=a.antenna,
        readout_path=a.readout,
        skychop_path=a.skychop,
        weather_path=a.weather,
        misti_path  =a.misti,
        cabin_path  =a.cabin,

        pixel_id   =a.pixel_id,
        coordinate =a.coordinate,
        mode       =a.mode,
        loadtype   =a.loadtype,
        findR      =a.findR,
        ch         =a.ch,
        Rth        =a.Rth,
        skyth      =a.skyth,
        cutnum     =a.cutnum,
        still      =a.still,
        period     =a.period,
        shuttle    =a.shuttle,
        lon_min_off=a.lon_min_off,
        lon_max_off=a.lon_max_off,
        lon_min_on =a.lon_min_on,
        lon_max_on =a.lon_max_on,
    )

    dems.to_zarr(a.filename, mode="w")

if __name__ == '__main__':
    main()



"""1つのKIDについて共振パラメータを計算する

Filename: calc_resonance_params.py
(C) 2021 内藤システムズ

使い方
------
コマンドラインで第一引数に一時ファイル(*.pkl)の名前を指定する。

例)
  $ python calc_resonance_params.py cache/kid00001.pkl
"""
import os
import sys
import pickle
from . import demerge

def calc_resonance_params(kid, nfwhm=5, minq=100, maxratio=1.0):
    """1つのKIDに対するresonanceパラメータを求める
    引数
    ----
    dict  KIDに関する測定値などをまとめた辞書(divide_data()関数によって作成される)
    float FWHMを単位としたフィット範囲[オプション]
    float qの最小値(フィットの良し悪しを判定する基準として利用する:demerge.find_peaks()関数で利用)[オプション]
    float フィットの良し悪しを判定する基準((bg-d)/bgの最大値:demerge.find_peaks()関数で利用)[オプション]

    戻り値
    ------
    dict フィットの結果を引数で与えられたkidに追加した辞書
    """
    fc = kid['tod'].frequency
    peaks = demerge.find_peaks(kid['localsweep'].x, kid['localsweep'].amplitude, fc, minq=minq, maxratio=maxratio)
    result, fitrange = demerge.fit_onepeak(kid['localsweep'], peaks, nfwhm=nfwhm)

    fr = result.params['fr'].value # GHz
    dfr = result.params['fr'].stderr # GHz
    Qr = result.params['Qr'].value
    dQr = result.params['Qr'].stderr
    Qc = result.params['Qc'].value
    dQc = result.params['Qc'].stderr
    Qi = result.params['Qi'].value
    dQi = result.params['Qi'].stderr
    if dfr/fr<0. or dQr/Qr<0. or dQc/Qc<0. or dQi/Qi<0.:
        kid['enabled'] = False
    elif fr!=fr or dfr!=dfr or Qr!=Qr or dQr!=dQr: # reject Nan
        kid['enabled'] = False
    elif abs(Qr)==float('inf') or abs(dQr)==float('inf'): # reject +/-inf
        kid['enabled'] = False
    if kid['enabled']:
        kid['resonance_params'] = result.params
        kid['fitrange'] = fitrange
    else:
        kid['resonance_params'] = None
        kid['fitrange'] = None
    return kid

def main() -> None:
    """
    コマンドライン引数
    ------------------
    args[1] string divide_data()関数で生成したpklファイルへのパス
    """
    args = sys.argv
    filename = args[1]
    with open(filename, 'rb') as f:
        kid = pickle.load(f)
    kid = calc_resonance_params(kid)
    with open(filename, 'wb') as f:
        pickle.dump(kid, f)

if __name__=='__main__':
    main()



"""DESHIMA解析プログラム

File name: demerge.py
Python 3.7
(C) 2021 内藤システムズ
"""
import os
import sys
from warnings import catch_warnings, simplefilter, warn
import numpy as np
import scipy
import scipy.signal
import json
import pickle
import lmfit
import sympy
from collections.abc import Mapping
import astropy.io.fits as pyfits
from astropy.io import fits

#-------------------------
# For make_divided_data.py
def divide_data(kidslist, localsweep, tods):
    """KID毎にSweepデータやTODを分割して辞書に格納する
    引数
    ----
    list demerge.load_kidslist()関数で得られるタプル(info, kid_bins, blind_bins, powers)
    list SweepDataオブジェクトのリスト
    TODs TODsオブジェクト

    戻り値
    ------
    リスト
    """
    result = []
    for i, (nbin, l, r) in enumerate(find_nearest_blind_tones(kidslist)):
        result.append({
            'pixelid': 0,
            'kidid': i,
            'bin': nbin,
            'fftgain': tods.fftgain,
            'framert': tods.framert,
            'nbins': tods.nbins,
            'npoints': tods.npoints,
            'lofreq': tods.lofreq,
            'localsweep': localsweep[nbin],
            'readpower': kidslist[3][nbin],
            'blind_tone_left': tods[l], #FixedData class
            'blind_tone_right': tods[r], #FixedData class
            'tod': tods[nbin], #FixedData class
            'tod_timestamp': tods.timestamp,
            'enabled': True,
        })
    return result

def find_nearest_blind_tones(kidslist):
    """ KID信号の左右にある最も近いblind信号を探す
    引数
    ----
    kidslist(demerge.read_kidslist()関数の戻り値)

    戻り値
    ------
    list [[KID信号, 左blind tone, 右blind tone], ...]

    戻り値の例
    ----------
    [[-196483 -201203 -192949]
     [-189634 -192949 -184695]
     [-188948 -192949 -184695]
    ...
     [ 154774  153170  161962]
     [ 158692  153170  161962]
     [ 169653  161962  171230]]

    詳細
    ----
    この関数は引数に与えられたkidslistに含まれるkidsとblindsを利用する。
    kidsとblindsを組み合わせて各kid信号の両側にある最も近いblind信号を探す。
    見つかった左右のblind信号とkid信号をリストにまとめて返す。戻り値はリストの入れ子構造となる。
    blind信号が見つからない場合はNoneが格納される。
    戻り値はkidのbin番号で昇順になっている。
    この関数はbbsweeplib.TODcalibに定義されていたkids_with_both_blinds()関数を単機能にしたもの。

    探索の方法
    ----------
    kid信号とblind信号を合わせたリストを作成する。このリストはbin番号で昇順に並べ替えられている。
    あるkid信号のindexを取得し、そこを起点に左右に探索する(indexを減少または増加させる)。起点は含まない。
    見つかった信号のbin番号がblind信号のリストに含まれる場合は、それをそのkid信号のblind信号とする。
    """
    info, kids, blinds, powers = kidslist
    all_tones = sorted(list(kids) + list(blinds))
    result = []
    for kid in sorted(kids):
        left_tone = None
        i = all_tones.index(kid) - 1
        while i >= 0:
            tone = all_tones[i]
            if tone in blinds:
                left_tone = tone
                break
            i -= 1
        right_tone = None
        i = all_tones.index(kid) + 1
        while i < len(all_tones):
            tone = all_tones[i]
            if tone in blinds:
                right_tone = tone
                break
            i += 1
        result.append([kid, left_tone, right_tone])
    return result

def load_kidslist(filename):
    """kids.listファイルを読み込みBIN番号やローカル発振器の周波数などを取得する
    引数
    ----
    kids.listファイルへのパス

    戻り値
    ------
    dict ローカル発振器の周波数とフレーム長を辞書にまとめたもの
    list KID信号のBIN番号のリスト
    list blind信号のBIN番号のリスト
    dict BIN番号をキー(Int型)にしたpower(float型)の辞書

    詳細
    ----
    kids.listファイルの最初の2行はヘッダーとして処理する。
    1行目はローカル発振器の周波数、2行目はフレーム長である信じて読み込む。
    3行目以降は「#KIDs」か「#blinds」かBIN番号とpowerのペアが記されているものとする。
    「#KIDs」があった場合はそれ以降のBIN番号はkid_bins変数のリストに追加する。
    「#blinds」があった場合はそれ以降のBIN番号はblind_bins変数のリストに追加する。
    bins変数はkid_binsリストへの参照かblind_binsリストへの参照かになる。
    bins変数によってmode変数やフラグ変数による判定が必要なくなる。

    読み込むkids.listファイルの形式
    -------------------------------
    #LO:5600.0000
    #framelen:19
    #KIDs
    -196483 -75
    -189634 -75
    -188948 -75
        :
        :
    154774  -75
    158692  -78
    169653  -78
    #blinds
    -232027 -75
    -224371 -75
    -216715 -75
        :
        :

    注意
    ----
    ファイルから読み取った1行(line)は必ずstrip()で処理しないと正しく読み込めない。
    bins変数とnbin変数を混同しないように注意されたい。
    bins変数はkid_binsリストかblind_binsリストへの参照であり、nbin変数はBIN番号を一時的に格納する変数である。
    """
    MHz = 1e6
    info = {}
    powers = {}
    kid_bins = []
    blind_bins = []
    with open(filename, 'rt') as f:
        # 最初の2行はヘッダーとして処理する
        line = f.readline().strip() # 行末の空白や改行などを削除する
        name, lofreq = line.split(':')
        info[name[1:]] = float(lofreq)*MHz # name変数に格納されている文字列から最初の"#"を除いた部分をキーとする

        line = f.readline().strip()
        name, framelen = line.split(':')
        info[name[1:]] = int(framelen)

        bins = None # kid_binsリストかblind_binsリストかへの参照になる
        for line in f:
            line = line.strip()
            if line[1:] == 'KIDs':
                bins = kid_bins
                continue
            elif line[1:] == 'blinds':
                bins = blind_bins
                continue
            nbin, power = line.split()
            bins.append(int(nbin))
            powers[int(nbin)] = float(power)
    return info, kid_bins, blind_bins, powers

def load_localsweep(filename, framelen):
    """ローカルsweepデータを読み込みBIN番号とSweepDataオブジェクトを対応づけた辞書として返す
    引数
    ----
    string  ローカルsweepファイルへのパス
    integer フレーム長

    戻り値
    ------
    dict BIN番号とSweepDataオブジェクト対応づけた辞書

    説明
    ----
    元のプログラムではread_localsweep()関数と_read_sweep()関数に分かれていた処理を一緒にした。
    read_localsweep()関数ではOrderedDict()を用いてSweepDataをBIN番号順に並べ替えていたが、
    この後の計算や処理でBIN番号順に並んでいる必要がないので、単純な辞書にSweepDataを格納することとした。
    並べ替えなどの処理を省いた。
    またSweepDataの辞書を作る際に辞書包括表記を用いた。

    参考文献
    --------
    spaceKIDS Readout System Manula(SK-SRON-018-V0.3) page.7
    上記のページにローカルsweepデータの読み方と各種計算方法が記されている。
    """
    MHz = 1e6
    sample_rate = 2e9 # Sample/sec
    rawdata = np.loadtxt(filename)
    bins    = map(int, rawdata[0, 1::3]) #リスト内包表記よりもmap()関数の方が速い
    idata   = rawdata[:, 2::3].T
    qdata   = rawdata[:, 3::3].T
    lofreqs = rawdata[:, 0]*MHz
    dfreq   = sample_rate/(2**framelen) # frequency resolution
    return {b: SweepData(i, q, lofreqs + b*dfreq) for b, i, q in zip(bins, idata, qdata)}

#-----------------------------
# For calc_resonance_params.py
def fit_onepeak(sweepdata, peaks, nfwhm=5):
    """Sweepデータをgaolinbgでフィットする
    引数
    ----
    SweepData Sweepデータオブジェクト
    list      find_peaks()関数で得られる各ピークのパラメタのリスト
    integer   nfwhm

    戻り値
    ------
    MinimizerResult gaolinbgによるフィットの結果
    slice           フィット範囲(インデックス)
    """
    initial_params = gaolinbg_guess(sweepdata, peaks) #params = guess(data)
    gaolinbg_param_names = initial_params.keys()
    s = slice(None)
    if nfwhm > 0:
        if len(peaks) == 1:
            pdict = peaks[0]
        else:
            raise Exception("number of peak not 1: %d" % len(peaks))
        s = adjust_fitrange(nfwhm, len(sweepdata), len(gaolinbg_param_names), pdict)

    # gaolinbg関数をパラメタで全微分した関数を取得
    Dfun = gaolinbg_param_gradient_function()

    def wrapped_Dfun(x, arga, absa, tau, fr, Qr, Qc, phi0, c):
        """全微分の導関数の結果を2次元ベクトルに変換する
        説明
        ----
        lmfit.Minimize()関数では複素関数を扱えない為、このような変換を行う。
        """
        return complex_to_cartesian2darray(Dfun(x, arga, absa, tau, fr, Qr, Qc, phi0, c))

    def Dfun_pars(p, x, dat=None):
        v = dict((k, v) for (k, v) in p.valuesdict().items() if k in gaolinbg_param_names)
        return wrapped_Dfun(x, **v)

    kws = {}
    kws['Dfun'] = Dfun_pars
    kws['col_deriv'] = 1

    gaolinbg_func = gaolinbg_function()
    ys = complex_to_cartesian2darray(sweepdata.iq[s])
    def residue(p, x):
        """データとの差を計算する
        引数
        ----
        lmfit.Parameters
        numpy.array 周波数

        戻り値
        ------
        lmfit.MinimizerResult フィットの結果

        注意
        ----
        lmfit.minimize()は複素数の最小化はできない。
        なので、複素数を一旦2次元ベクトルに変換して計算する。
        """
        v = {k: v for (k, v) in p.valuesdict().items() if k in gaolinbg_param_names}
        return complex_to_cartesian2darray(gaolinbg_func(x, **v)) - ys

    with catch_warnings():
        simplefilter("ignore")
        params = lmfit.Parameters()
        params.add('arga', initial_params['arga'])
        params.add('absa', initial_params['absa'])
        params.add('tau',  initial_params['tau'])
        params.add('fr',   initial_params['fr'])
        params.add('Qr',   initial_params['Qr'])
        params.add('Qc',   initial_params['Qc'])
        params.add('phi0', initial_params['phi0'])
        params.add('c',    initial_params['c'])
        params.add('Qi',   expr='1/(1/Qr - 1/Qc*cos(phi0))')
        minimizer_result = lmfit.minimize(residue, params, args=(sweepdata.x[s],), **kws)
        return minimizer_result, s # fit結果, フィット範囲

def search_peak(sweepdata, peaks):
    """複数のピークが見つかった場合は中心周波数に一番近いピークを採用する
    引数
    ----
    SweepData Sweepデータオブジェクト
    list      find_peaks()関数によって得られる各ピークのパラメタをまとめたリスト

    戻り値
    ------
    dict 選択されたピークのパラメタをまとめた辞書
    """
    center  = (sweepdata.x[0] + sweepdata.x[-1])/2.0
    minind  = 0
    mindist = abs(peaks[0]['f0'] - center)
    for i in range(1, len(peaks)):
        if mindist > abs(peaks[i]['f0'] - center):
            minind  = i
            mindist = abs(peaks[i]['f0'] - center)
    return peaks[minind]

def find_peaks(freq, ampl, fc, smooth=None, threshold=None, minq=10000, maxratio=0.5):
    """Search peaks in data.
    引数
    ----
    freq      周波数の配列
    ampl      amplitudeの配列
    fc        TODのキャリア周波数
    smooth    平滑化のwindow長(正の整数)[オプション]
    threshold ピークの判定基準[オプション]
    minq      Qの最小値(ピークの良し悪しの判定に利用する)[オプション]
    maxratio  (bg - d)/bgの最小値(ピークの良し悪しの判定に利用する)[オプション]

    戻り値
    ------
    list Lorentzianのフィット共にgaolinbgのパラメタの初期値を決定するのに必要なパラメタを辞書にまとめたリスト
    """
    dfreq = freq[1] - freq[0] # frequency resolution
    if smooth is None:
        smooth = 15

    # 平滑化
    deriv2 = scipy.signal.savgol_filter(ampl, oddify(smooth), 3, 2, delta=dfreq)
    ampl_s = scipy.signal.savgol_filter(ampl, oddify(smooth), 3, 0, delta=dfreq)

    # ピーク探索
    if threshold is None:
        d2sigma   = np.std(deriv2)
        d2sigma   = np.std(deriv2[np.abs(deriv2)<2*d2sigma])
        threshold = 3*d2sigma
    kid_indices = []
    for i in np.where(deriv2 > threshold)[0]:
        if i < len(deriv2) - 1 and deriv2[i - 1] <= deriv2[i] and deriv2[i] >= deriv2[i + 1]:
            kid_indices.append(i)

    kids = [] # 各KIDをLorentzianでフィットした結果(Lorentzianのパラメタ)
    nbadq = 0 # Qが条件を満たさなかったKIDの数
    nbadd = 0 # maxratioを満たさなかったKIDの数
    if not kid_indices:
        return []

    for i in kid_indices:
        #
        # 詳細
        # ----
        # ピークの左右でamplitudeの2階の微係数が0になる部分(変曲点)を利用してフィット範囲を設定する。
        # フィットの結果Qと(bg-d)/bgが条件を満たさない場合は、そのピークは無視される。
        #
        l, r = i, i
        while l > 0           and deriv2[l] > 0: l -= 1
        while r < len(deriv2) and deriv2[r] > 0: r += 1
        w = (r - l + 1)*dfreq
        w = w*6.0/np.sqrt(3) # convert to FWHM
        l = int((l - i)*6.0/np.sqrt(3) + i)
        r = int((r - i)*6.0/np.sqrt(3) + i)
        if l < 0         : l = 0
        if r >= len(freq): r = len(freq) - 1

        q0 = freq[i]/w
        f, q, d, bg = fitLorentzian(freq[l:r], ampl[l:r], freq[i], q0)
        ##### refitting by extending fit range
        if (bg - d)/bg > maxratio:
            n = 10
            if l - n >= 0 and r + n < len(freq):
                f, q, d, bg = fitLorentzian(freq[l-n:r+n], ampl[l-n:r+n], freq[i], q0)
                print('rough refitting with Lorentzian...')
        if q < minq:
            nbadq += 1
            continue
        if (bg - d)/bg > maxratio:
            nbadd += 1
            continue
        kids.append((f, q, d, bg))
    del l, r, f, q, d, bg
    if nbadq > 0:
        print('removed', nbadq, 'peaks with bad Q')
    if nbadd > 0:
        print('removed', nbadd, 'peaks with bad S21min')
    kids.sort() #sort by frequency

    # pick up a peak which is closest to the carrier frequency (when fc in freq range)
    if len(kids) > 0 and fc >= freq[0] and fc <= freq[-1]:
        i = np.argmin(abs(np.array(kids).T[0] - fc))
        kids = [kids[i]]

    # gaolinbgのパラメタの初期値を決めるためのbackgroundなどを計算する
    for i, (f, q, depth, bg) in enumerate(kids):
        f0ind = np.argmin(abs(freq - f))
        w     = f/q
        dl    = w/2.0/dfreq
        dr    = w/2.0/dfreq
        bg_l  = ampl_s[max(int(f0ind - 3*dl), 0)]
        if int(f0ind - 3*dl)<0:
            bg_r = ampl_s[0]
        else:
            bg_r  = ampl_s[min(int(f0ind - 3*dl), len(freq) - 1)]
        a_off = (bg_l + bg_r)/2.0
        a_on  = ampl_s[f0ind]
        kids[i] = {
            'Q': q,
            'f0': f,
            'f0ind': f0ind,
            'dl': int(dl),
            'dr': int(dr),
            'a_off': a_off,
            'a_on': a_on,
            'ampl': depth,
            'bg': bg
        }
    return kids

def oddify(n):
    """n/2を切り上げそれを2倍して1足し奇数にして返す
    引数
    ----
    integer

    戻り値
    ------
    integer 奇数
    """
    return int(np.ceil(n/2.0)*2+1)

def fitLorentzian(freq, ampl, f0, q0):
    """Fit data with Lorentzian curve.
    引数
    ----
    freq a 1-D array of frequency
    ampl a 1-D array of amplitude
    f0   initial parameter for center frequency
    q0   initial parameter for quality factor

    戻り値
    ------
    (fc, q, d, bg) フィットの結果をまとめたタプル

    パラメタの説明
    --------------
    f : center frequency
    q : quality factor
    d : amplitude for Lorentzian curve
    bg: constant background level
    """
    def f(x):
        (a, b, c, d) = x # background, amplitude, 2/FWHM, freq. center
        y = a + b / (((freq-d)*c)**2 + 1)
        return y - ampl
    a = np.median(ampl)
    b = -0.8 * a
    c = 2.0 * q0 / f0
    d = f0
    x0 = np.array([a, b, c, d])
    x1 = scipy.optimize.leastsq(f, x0)[0]
    (a, b, c, d) = x1
    fc = d
    q = abs(c * d / 2.0)
    return (fc, q, -b, a)

def adjust_fitrange(nfwhm, ndata, nparams, peakparams, factor=1):
    """Adjust fit range to make sure :math:`N_{free} \geq nparam`
    引数
    ----
    float   nfwhm: minimum fit length in unit of FWHM
    integer ndata: length of data
    integer nparams: number of parameter
    dict    peakparams: peak information
    integer factor

    戻り値
    ------
    tuple (rbegin, rend) フィット範囲の始まりと終わりのインデックス

    factorについて
    --------------
    Specify 2 as 'factor' if you fit complex variable(as 2d-vector).

    利用するピークパラメタ
    ----------------------
     - f0ind: a integer, index of peak center
     - dl   : a integer, index count to reach the left half-maximum
     - dr   : a integer, index count to reach the right half-maximum
    """
    f0ind = peakparams['f0ind']
    l, c, r = f0ind-peakparams['dl'], f0ind, f0ind+peakparams['dr']
    if (r - l + 1)*factor >= nparams:
        rbegin, rend = int(c-nfwhm*(c-l)), int(c+nfwhm*(r-c))
    else:
        n = nparams/(float(r - l + 1)*factor)
        rbegin, rend = int(f0ind-nfwhm*n*(c-l)), int(c+nfwhm*n*(r-c))
    if rbegin < 0:
        if rend + (-rbegin) >= ndata:
            rend = ndata-1
            rbegin = 0
        else:
            rend = rend + (-rbegin)
            rbegin = 0
    if rend >= ndata:
        if rbegin - (rend - ndata) < 0:
           rbegin = 0
           rend = ndata-1
        else:
           rbegin = rbegin - (rend - ndata)
           rend = ndata-1
    if (rend - rbegin + 1)*factor < 11:
        raise Exception("Fit range guess error")
    return slice(rbegin, rend)

#-------------------------
# For make_reduced_fits.py
def readout_dict():
    hdr_key_lis = ['EXTNAME', 'FILENAME', 'FRAMERT', 'FRAMELEN', 'DSAMPLE',]
    hdr_val_lis = ['READOUT', None, None, None, None]
    hdr_com_lis = ['name of binary data',
                   'input filename',
                   'sampling rate',
                   '2-log of frame length',
                   'number of down sampling',
                   'label for field 0', 'data format of field 0',
                   'label for field 1', 'data format of field 1',]
    cols_key_lis = ['timestamp', 'pixelid',]
    cols_data_lis = []
    tform = ['D', 'I',]
    tunit = [None, None,]

    r_dict = {'hdr_key_lis': hdr_key_lis,
              'hdr_val_lis': hdr_val_lis,
              'hdr_com_lis': hdr_com_lis,
              'cols_key_lis': cols_key_lis,
              'cols_data_lis': cols_data_lis,
              'tform': tform,
              'tunit': tunit}
    return r_dict

def kids_dict():
    hdr_key_lis = ['EXTNAME', 'FILENAME',]
    hdr_val_lis = ['KIDSINFO', None,]
    hdr_com_lis = ['name of binary data',
                   'localsweep filename',
                   'label for field 0', 'data format of field 0',
                   'label for field 1', 'data format of field 1',
                   'label for field 2', 'data format of field 2', 'data unit of field 2',
                   'label for field 3', 'data format of field 3', 'data unit of field 3',
                   'label for field 4', 'data format of field 4',
                   'label for field 5', 'data format of field 5', 'data unit of field 5',
                   'label for field 6', 'data format of field 6',
                   'label for field 7', 'data format of field 7',
                   'label for field 8', 'data format of field 8']
    cols_key_lis = ['pixelid', 'kidid', 'Pread', 'fc', 'yfc, linyfc',
                    'fr, dfr (300K)', 'Qr, dQr (300K)', 'Qc, dQc (300K)', 'Qi, dQi (300K)']
    cols_data_lis = []
    tform = ['I', 'I', 'E', 'E', '2E', '2E', '2E', '2E', '2E']
    tunit = [None, None, 'dBm', 'GHz', None, 'GHz', None, None, None]

    k_dict = {'hdr_key_lis': hdr_key_lis,
              'hdr_val_lis': hdr_val_lis,
              'hdr_com_lis': hdr_com_lis,
              'cols_key_lis': cols_key_lis,
              'cols_data_lis': cols_data_lis,
              'tform': tform,
              'tunit': tunit}
    return k_dict

def rebin_array(inarray, rebin):
    return inarray[:(inarray.size//rebin)*rebin].reshape(-1, rebin).mean(axis=1)

def find_glitch(yss,
                baseline_thresh = 6.0, glitch_thresh = 5.0, clusterize_thresh = 2, offset = 0,
                opt_diff2=True, smrange=1, returnval=False):
    """
    find glitches common to `yss`, assuming glitch exists at the same
    time of all data in `yss`. Too close glitches or broad glitch are
    treated as one glitch.

    引数
    ----
    yss              : an array of 1-D arrays of data to deglitch
    baseline_thresh  : threshold to decide baseline
    glitch_thresh    : threshold to decide as glitch
    clusterize_thresh: if gap between glitches are less than or equal to this, treat them as one glitch.
    offset           : an integer offset added to returned bad array
    opt_diff2        : bool to select whether to use 2nd derivative for glitch identification
    smrange          : an integer to set range of rolling mean (used for glitch identification based on std)
    returnval        : bool to choose whether to return threshold value or not

    戻り値(returnvalの値によって変化する)
    -------------------------------------
    an 1-D array of boolean, that can be used as index of ys (elements of yss).
    (option) threshold value for glitch identification (return when returnval=True)
    """
    ave = np.average(yss, axis=0)
    xs  = np.arange(len(yss[0]))
    dx  = xs[1] - xs[0]
    if opt_diff2:
        diff2 = np.array( _numdif_2(ave, dx) )
        sigma = np.std(diff2)
        good  = (np.abs(diff2) < (baseline_thresh*sigma))
        sigma = np.std(diff2[good])
        bad   = (np.abs(diff2) >= (glitch_thresh*sigma))
        if returnval:
            thre = glitch_thresh*sigma/np.sqrt(6)*(dx**2)
    else:
        ##### rolling mean
        weights = np.ones(smrange)/smrange
        assert( len(ave)>len(weights) )
        smooth  = np.convolve(ave, weights, 'same')

        invalid = int( (len(weights)-1)/2 )
        smooth[:invalid] = np.mean(ave[:invalid]) * np.ones( len(ave[:invalid]) )
        smooth[-invalid-1:] = np.mean(ave[-invalid-1:]) * np.ones( len(ave[-invalid-1:]) )

        ave_sm = ave - smooth

        mean  = np.mean(ave_sm)
        sigma = np.std(ave_sm)
        good  = (np.abs(ave_sm-mean) < (baseline_thresh*sigma))
        mean  = np.mean(ave_sm[good])
        sigma = np.std(ave_sm[good])
        bad   = (np.abs(ave_sm-mean) >= (glitch_thresh*sigma))
        if returnval:
            thre = glitch_thresh*sigma
    ## treat broad glitch (or too close glitches) as one glitch
    bad = _clusterize_indices(bad, clusterize_thresh)
    if opt_diff2:
        bad = np.concatenate( ([bad[0]], bad, [bad[-1]]) )

    bad_ = np.copy(bad)
    if offset>0:
        for i,b in enumerate(bad):
            if b:
                bad_[i] = True
                for j in range(offset):
                    if (i+j+1)<len(bad): bad_[i+j+1] = True
    if returnval:
        return bad_, thre
    else:
        return bad_

def _numdif_2(y, dx):
    return (y[2:]+y[:-2]-2*y[1:-1])/dx**2

def _clusterize_indices(indices, threshold):
    """Fill small gap (within `threshold`) in indices.
    引数
    ----
    indices:   an 1-D array of Boolean
    threshold: an interger, allowed gap between True's in indices

    戻り値
    ------
    list 真理値のリスト

    実行結果の例
    ------------
     If threshold == 1,
      [True, False, False, True, True] => [True, False, False, True, True].
     If threshold == 2,
      [True, False, False, True, True] => [True, True, True, True, True].
    """

    results = np.copy(indices)
    prev  = 0
    first = True
    for i, x in enumerate(results):
        if x:
            if (not first) and i - prev <= threshold + 1:
                for j in range(prev, i):
                    results[j] = True
            prev  = i
            first = False
    return results

def calibrate_with_blind_tones(signal, blind_left, blind_right):
    """do blind tone calibration.
    引数
    ----
    signal     : signal tone TOD.
    blind_left : left blind tone TOD.
    blind_right: right blind tone TOD.

    戻り値
    ------
    a 1-D array of complex value (calibrated IQ)

    詳細
    ----
    each parameter is either a FixedData or a tuple (freq, IQ)
    where freq is a frequency [Hz],
    IQ is a 1-D array of IQ data as complex value.
    """
    def get_f_iq(fx):
        if isinstance(blind_left, FixedData):
            return fx._frequency, fx.iq
        else:
            return fx
    f_s, iq_s = get_f_iq(signal)
    f_l, iq_l = get_f_iq(blind_left)
    f_r, iq_r = get_f_iq(blind_right)

    calib_l = iq_l / np.average(iq_l)
    calib_r = iq_r / np.average(iq_r)
    calib_s = calib_l + (calib_r - calib_l)/(f_r - f_l) * (f_s - f_l) # interpolate

    calibrated = (iq_s / calib_s)
    return calibrated

def interpolate_bad(ys, bad):
    """Linear interpolate `ys` in region `bad`.
    引数
    ----
    ys : an 1-D array
    bad: an 1-D array of boolean, with True indicating bad region in `ys`.

    戻り値
    ------
    an 1-D array, with bad region interpolated.
    """
    xs  = np.arange(len(ys))
    deglitched = np.interp(xs, xs[~bad], ys[~bad])
    return deglitched

def createBinTableHDU(data_dict):
    #-------- Set Header and Comments
    header = fits.Header()
    for (i, j, k) in zip(data_dict['hdr_key_lis'], data_dict['hdr_val_lis'], data_dict['hdr_com_lis']):
        header[i] = j, k
    #-------- Create Collumns of the Binary Table
    columns = []
    for i in range(len(data_dict['cols_key_lis'])):
        columns.append(fits.Column(name=data_dict['cols_key_lis'][i],
                                    format=data_dict['tform'][i],
                                    array=data_dict['cols_data_lis'][i],
                                    unit=data_dict['tunit'][i]))
    hdu = fits.BinTableHDU.from_columns(columns, header)
    #-------- Add comments
    for i in range(-(len(data_dict['hdr_com_lis'])-data_dict['hdr_com_lis'].index('label for field 0')), 0):
        addHeaderComments(hdu.header, list(hdu.header)[i], data_dict['hdr_com_lis'][i])
    return hdu

def addHeaderComments(hdr, key, com):
    """Add Comments to Header"""
    hdr.comments[key] = com

def f_Lorentzian(params, freq, data=None):
    """Lorengzianモデル関数
    mkid_data/peak_search.pyで使われている関数と同じ。
    引数のdataが指定された場合はモデル関数の計算値とdataの差を返す。

    引数
    ----
    numpy.Parameters
    numpy.array 周波数
    numpy.array amplitude

    戻り値
    ------
    モデルの計算値|計算値とdataの差
    """
    params = params.valuesdict()
    b      = params['bg']
    a      = params['amp']
    fc     = params['f0']
    fwhm   = params['fwhm']
    model = b + a/(((freq - fc)*(2.0/fwhm))**2 + 1)
    if data is None:
        return model
    return model - data

def gaolinbg_guess(sweepdata, peaks):
    gaolinbg_paramnames = 'arga absa tau fr Qr Qc phi0 c'.split()
    x      = sweepdata.x
    deltax = x[1] - x[0]
    pdict  = search_peak(sweepdata, peaks)
    y0     = sweepdata.iq[pdict['f0ind']]
    FWHM   = pdict['f0']/pdict['Q']
    ddeg   = sweepdata.deg[1:] - sweepdata.deg[:-1]
    tau    = -np.average(ddeg[abs(ddeg)<180])*np.pi/180.0/deltax/2/np.pi
    f0     = pdict['f0']
    theta  = np.angle(y0)
    arga   = np.angle(y0*np.exp(1j*2*np.pi*tau*f0))
    absa   = pdict['a_off']
    fr     = f0
    Qr     = f0/FWHM
    Qc     = Qr
    phi0   = 0
    c      = 0
    return dict(zip(gaolinbg_paramnames, (arga, absa, tau, fr, Qr, Qc, phi0, c)))

def gaolinbg_rewind(x, y, arga, absa, tau, fr, Qr, Qc, phi0, c):
    tmp = y/absa/np.exp(-1j*(2*np.pi*x*tau - arga)) - c*(x - fr)
    return (tmp - 1)*Qc/Qr/np.exp(1j*phi0) + 0.5

def gaolinbg_expression():
    arg_symbols = sympy.symbols(r'(\arg{a}) |a| tau f_r Q_r Q_c phi_0 c')
    arg_names = 'arga absa tau fr Qr Qc phi0 c'.split()
    arga, absa, tau, fr, Qr, Qc, phi0, c = arg_symbols
    x = sympy.symbols('x')
    from sympy import exp, I, pi
    expr = (absa*exp(-I*(2*pi*x*tau - arga))*(1 + c*(x - fr) - Qr/Qc*exp(I*phi0)/(1 + 2*I*Qr*((x - fr)/fr))))
    return expr, arg_symbols, arg_names

def gaolinbg_bg_expression():
    arg_symbols = sympy.symbols(r'(\arg{a}) |a| tau f_r Q_r Q_c phi_0 c')
    arg_names = 'arga absa tau fr Qr Qc phi0 c'.split()
    arga, absa, tau, fr, Qr, Qc, phi0, c = arg_symbols
    x = sympy.symbols('x')
    from sympy import exp, I, pi
    expr = (absa*exp(-I*(2*pi*x*tau - arga))*(1 + c*(x - fr)))
    return expr, arg_symbols, arg_names

def gaolinbg_function():
    expr, arg_symbols, arg_names = gaolinbg_expression()
    expr_ = expr.subs(zip(arg_symbols, arg_names))
    args = sympy.symbols(['x'] + arg_names)
    return sympy.lambdify(args, expr_, 'numpy', dummify=False)

def gaolinbg_bg_function():
    expr, arg_symbols, arg_names = gaolinbg_bg_expression()
    expr_ = expr.subs(zip(arg_symbols, arg_names))
    args = sympy.symbols(['x'] + arg_names)
    return sympy.lambdify(args, expr_, 'numpy', dummify=False)

def gaolinbg_param_gradient_function():
    expr, arg_symbols, arg_names = gaolinbg_expression()
    funcargs = ['x'] + arg_names
    arga, absa, tau, fr, Qr, Qc, phi0, c = arg_symbols
    gradfuncs = [ #各パラメータに関する導関数を求め、関数化(Pythonで使える)する。
        sympy.lambdify(funcargs, sympy.diff(expr, arga).subs(zip(arg_symbols, arg_names)), 'numpy', dummify=False),
        sympy.lambdify(funcargs, sympy.diff(expr, absa).subs(zip(arg_symbols, arg_names)), 'numpy', dummify=False),
        sympy.lambdify(funcargs, sympy.diff(expr, tau ).subs(zip(arg_symbols, arg_names)), 'numpy', dummify=False),
        sympy.lambdify(funcargs, sympy.diff(expr, fr  ).subs(zip(arg_symbols, arg_names)), 'numpy', dummify=False),
        sympy.lambdify(funcargs, sympy.diff(expr, Qr  ).subs(zip(arg_symbols, arg_names)), 'numpy', dummify=False),
        sympy.lambdify(funcargs, sympy.diff(expr, Qc  ).subs(zip(arg_symbols, arg_names)), 'numpy', dummify=False),
        sympy.lambdify(funcargs, sympy.diff(expr, phi0).subs(zip(arg_symbols, arg_names)), 'numpy', dummify=False),
        sympy.lambdify(funcargs, sympy.diff(expr, c   ).subs(zip(arg_symbols, arg_names)), 'numpy', dummify=False),
    ]
    def total_derivative(x, arga, absa, tau, fr, Qr, Qc, phi0, c):
        return np.array(np.broadcast_arrays(
            gradfuncs[0](x, arga, absa, tau, fr, Qr, Qc, phi0, c),
            gradfuncs[1](x, arga, absa, tau, fr, Qr, Qc, phi0, c),
            gradfuncs[2](x, arga, absa, tau, fr, Qr, Qc, phi0, c),
            gradfuncs[3](x, arga, absa, tau, fr, Qr, Qc, phi0, c),
            gradfuncs[4](x, arga, absa, tau, fr, Qr, Qc, phi0, c),
            gradfuncs[5](x, arga, absa, tau, fr, Qr, Qc, phi0, c),
            gradfuncs[6](x, arga, absa, tau, fr, Qr, Qc, phi0, c),
            gradfuncs[7](x, arga, absa, tau, fr, Qr, Qc, phi0, c)))

    return total_derivative

def complex_to_cartesian2darray(x):
    x = np.atleast_1d(x)
    shape = x.shape
    return np.concatenate([np.real(x), np.imag(x)], axis=len(shape)-1)

def cartesian2darray_to_complex(x):
    assert x.shape[-1] % 2 == 0
    size = x.shape[-1] / 2
    return x[...,:size] + 1j*x[...,size:]

def power_spectrum_density(data, dt, ndivide=1, window=scipy.signal.windows.hann, overwrap_half=False):
    """Calculate power spectrum density of data.
    引数
    ----
    data         : input data
    dt           : time between each data
    ndivide      : do averaging(split data into ndivide, get psd of each, and average them)
    overwrap_half: a boolean, split data to half-overwrapped regions

    戻り値
    ------
    (frequencies, psd)
    """
    if overwrap_half:
        step = int(len(data)/(ndivide + 1))
        size = step*2
    else:
        step = int(len(data)/ndivide)
        size = step
    if bin(size).count("1") != 1:
        if overwrap_half:
            warn('warning: ((length of data)/(ndivide+1))*2 is not power of 2: %d' % size)
        else:
            warn('warning: (length of data)/ndivide is not power of 2: %d' % size)
    psd = np.zeros(size)
    T = (size - 1)*dt
    vs = 1/dt
    vk_ = scipy.fft.fftfreq(size, dt)
    vk = vk_[np.where(vk_ >= 0)]
    for i in range(ndivide):
        d = data[i*step:i*step+size]
        if window is None:
            w    = np.ones(size)
            corr = 1.0
        else:
            w    = window(size)
            corr = np.mean(w**2)
        psd  = psd + 2*(np.abs(scipy.fft.fft(d*w)))**2/size*dt/corr
    return vk, psd[:len(vk)]/ndivide

class SweepData():
    """mkid_data.data.SweepDataを単純化したクラス"""
    def __init__(self, i=None, q=None, freq=None):
        self.i = i
        self.q = q
        self.freq = freq
        return

    @property
    def x(self):
        GHz = 1e9
        return self.freq/GHz

    @property
    def amplitude(self):
        return np.sqrt(self.i**2 + self.q**2)

    @property
    def iq(self):
        return self.i + 1j*self.q

    @property
    def deg(self):
        return np.arctan2(self.q, self.i)*180/np.pi

    def __len__(self):
        return len(self.freq)

class FixedData():
    """mkid_data.data.FixedDataを単純化したクラス"""
    def __init__(self):
        self.timestamp = None
        self.freq = None
        self._frequency = None
        self._i = None
        self._q = None
        self.fftgain = None
        self.info = None
        self.rate = 1
        return

    @property
    def frequency(self):
        GHz = 1e9
        return self._frequency/GHz

    @property
    def iq(self):
        return self.i + 1j*self.q

    @property
    def i(self):
        return self._i/self.fftgain

    @property
    def q(self):
        return self._q/self.fftgain

    @property
    def t(self):
        return self.timestamp/self.rate

class FixedFitData():
    """mkid_data.data.FixedFitDataを単純化したクラス"""
    def __init__(self, freq=None, timestamp=None, ampl=None, phase=None, info=None):
        self._freq = freq
        self._timestamp = timestamp
        self._ampl = ampl
        self._phase = phase
        self._info = info
        return

    @property
    def frequency(self):
        GHz = 1e9
        return self._freq/GHz

    @property
    def amplitude(self):
        return self._ampl

    @property
    def phase(self):
        return self._phase

    @property
    def t(self):
        return self._timestamp

    @property
    def info(self):
        return self._info

    def unpack(self):
        return (self._timestamp, self._ampl, self._phase)

class TODs(Mapping):
    """TODデータを保持するクラス
    注意
    ----
    このクラスのインスタンスを作成すると自動的にfitsファイルが開かれる。
    TODに関する処理が終わったら必ずclose()メソッドを明示的に呼び出すこと。

    コピー元
    --------
    bbsweeplib.kids.py fits_tods_fits
    """
    chunksize = int(1e6)
    kSampleRate = 2e9
    def __init__(self, infile):
        """コンストラクタ
        引数
        ----
        string TODが保存されたfitsファイルの名前

        戻り値
        ------
        なし
        """
        self.infile  = infile
        self.open()

    def open(self):
        self.hud = pyfits.open(self.infile)
        bintable = self.hud[1]
        self.fftgain  = bintable.header['fftgain']
        self.framert  = bintable.header['framert']
        self.nbins    = bintable.header['nbins']
        self.npoints  = bintable.header['npoints']
        self.lofreq   = bintable.header['lofreq']
        self.bins     = [_to_nbit_signed(bintable.header['BIN%d' % i], self.npoints)
                         for i in range(self.nbins)]
        self.if_freq      = self.kSampleRate * np.array(self.bins) / 2**self.npoints
        self.carrier_freq = self.if_freq + self.lofreq

        bindata = bintable.data
        timestamp_ = bindata.field('timestamp')
        framenr_ = bindata.field('framenr')
        self.timestamp = timestamp_
        self.framenr = framenr_
        self._read_bins = {}

        self.offset = -100 # remove last part of data (sometimes they behave bad)

    def close(self):
        self.hud.close()

    def __len__(self):
        return len(self.bins)

    def __contains__(self, key):
        return key in self.bins

    def __iter__(self):
        for key in sorted(self.bins):
            yield key

    def __getitem__(self, key):
        if key in self._read_bins:
            return self._read_bins[key]
        else:
            ind = self.bins.index(key)
            chunksize=int(1e4)
            name = 'kid_%04d' % ind
            rawdata = self.hud[1].data['data']
            read_I_ = rawdata[:self.offset, 2*ind    ]
            read_Q_ = rawdata[:self.offset, 2*ind + 1]

            info = dict()
            info['bins'] = self.bins
            info['freqs'] = self.carrier_freq[ind]
            info['header'] = self.hud[1].header
            d = FixedData()
            d.timestamp = self.timestamp[:self.offset]
            d._i = read_I_
            d._q = read_Q_
            d.fftgain = self.fftgain
            d._frequency = self.carrier_freq[ind]
            d.info = info
            self._read_bins[key] = d
            return d

    def __getstate__(self):
        if hasattr(self, 'hud'):
            dic = self.__dict__.copy()
            dic['_read_bins'] = {}
            dic['hud'] = None
            return dic
        else:
            return self.__dict__

    def __setstate__(self, dic):
        self.__dict__ = dic.copy()
        if 'hud' in dic:
            self.open()

def _to_nbit_signed(x, n):
    if x > 2**(n-1):
        return -((~x & (2**n-1))+1)
    else:
        return x



"""1つのKIDについてSweepのフィットの結果をプロットする。

Filename: plot.py
(C) 2021 内藤システムズ

使い方
------
第一引数 フィットの結果が格納された一時ファイル(*.pkl)の名前
第二引数 グラフを保存するディレクトリ名

例)
  $ python plot.py cache/kid00001.pkl graph
"""
import os
import sys
import pickle
import matplotlib.pyplot as plt
import numpy as np
from scipy import interpolate
from . import demerge

def plot(kid, oldkid=None):
    if kid['enabled'] == False:
        return None, None
    if kid['resonance_params'] is None:
        return None, None
    arga = kid['resonance_params']['arga']
    absa = kid['resonance_params']['absa']
    tau = kid['resonance_params']['tau']
    fr = kid['resonance_params']['fr'].value # GHz
    Qr = kid['resonance_params']['Qr'].value
    Qc = kid['resonance_params']['Qc'].value
    phi0 = kid['resonance_params']['phi0']
    c = kid['resonance_params']['c']

    fitrange = kid['fitrange']
    rewound_localsweep = demerge.gaolinbg_rewind(kid['localsweep'].x, kid['localsweep'].iq, arga, absa, tau, fr, Qr, Qc, phi0, c)
    fitted_localsweep = demerge.gaolinbg_function()(kid['localsweep'].x, arga, absa, tau, fr, Qr, Qc, phi0, c)
    limited_fitted_localsweep = demerge.gaolinbg_function()(kid['localsweep'].x[fitrange], arga, absa, tau, fr, Qr, Qc, phi0, c)
    rewound_fitted_localsweep = demerge.gaolinbg_rewind(kid['localsweep'].x, fitted_localsweep, arga, absa, tau, fr, Qr, Qc, phi0, c)
    fitted_background = demerge.gaolinbg_bg_function()(kid['localsweep'].x, arga, absa, tau, fr, Qr, Qc, phi0, c)
    limited_fitted_background = demerge.gaolinbg_bg_function()(kid['localsweep'].x[fitrange], arga, absa, tau, fr, Qr, Qc, phi0, c)
    rewound_tod = demerge.gaolinbg_rewind(kid['tod'].frequency, kid['tod'].iq[::100], arga, absa, tau, fr, Qr, Qc, phi0, c)
    fitted_peak = demerge.gaolinbg_function()(fr, arga, absa, tau, fr, Qr, Qc, phi0, c)

    plt.ioff()
    fig1 = plt.figure(figsize=(16,10))
    ax1 = plt.subplot(231)
    ax2 = plt.subplot(232)
    ax3 = plt.subplot(233)
    ax4 = plt.subplot(212)

    # Sweepのプロット(KidFitResult::plot()から抜粋)
    ax1.plot(kid['localsweep'].x, abs(kid['localsweep'].iq),'.g', label='Ampl.')
    ax1.plot(kid['localsweep'].x, kid['localsweep'].i,'.b', label='I')
    ax1.plot(kid['localsweep'].x, kid['localsweep'].q,'.r', label='Q')
    ax1.plot(kid['localsweep'].x, abs(fitted_background), '-', color='gray', label='bg')
    ax1.plot(kid['localsweep'].x[fitrange], abs(limited_fitted_localsweep),'-y')
    ax1.plot(kid['localsweep'].x[fitrange], np.real(limited_fitted_localsweep),'-c')
    ax1.plot(kid['localsweep'].x[fitrange], np.imag(limited_fitted_localsweep),'-m')
    ax1.plot(fr, np.abs(fitted_peak),'r*')
    ax1.set_title('Amplitude vs Freq')
    ax1.set_xlabel('Frequency [GHz]')
    ax1.set_ylabel('Amplitude')
    ax1.axhline(color='k')
    ax1.legend(loc='best')
    ax1.grid()

    #KidFitResult::plot()から抜粋
    ax2.plot(kid['localsweep'].i, kid['localsweep'].q, '.b', label='data')
    ax2.plot(np.real(limited_fitted_localsweep), np.imag(limited_fitted_localsweep), '-c', label='fit')
    ax2.plot(np.real(fitted_peak), np.imag(fitted_peak), 'r*')
    ax2.plot(np.real(fitted_background), np.imag(fitted_background), '-', color='gray', label='bg')
    if not oldkid is None:
        ax2.plot(oldkid['I'], oldkid['Q'], '.b', label='data')
        ax2.plot(np.real(oldkid['fity']), np.imag(oldkid['fity']), '-c', label='fit')
        ax2.plot(np.real(fitted_peak), np.imag(fitted_peak), 'r*')
        ax2.plot(np.real(oldkid['bg']), np.imag(oldkid['bg']), '-', color='gray', label='bg')

    ax2.set_title('KID[{}], I vs Q'.format(kid['kidid']))
    ax2.set_xlabel('I')
    ax2.set_ylabel('Q')
    ax2.axhline(color='k')
    ax2.axvline(color='k')
    ax2.legend(loc='best')
    ax2.axis('equal')
    ax2.grid()

    #FitSweep()から抜粋
    ax3.set_title('rewind')
    ax3.plot(np.real(rewound_localsweep), np.imag(rewound_localsweep), label='Sweep')
    ax3.plot(np.real(rewound_tod), np.imag(rewound_tod), '.', alpha=0.2, label='TOD')
    ax3.plot(np.real(rewound_fitted_localsweep), np.imag(rewound_fitted_localsweep), 'y', lw=3, alpha=0.5, label='fit')
    ax3.set_xlabel('Re')
    ax3.set_ylabel('Im')
    ax3.axhline(color='k')
    ax3.axvline(color='k')
    ax3.legend(loc='best')
    ax3.axis('equal')
    ax3.grid()

    # TODのプロット
    phase_tod = -np.angle(-rewound_tod)
    phase_localsweep = -np.angle(-rewound_localsweep)
    phase_fitted_localsweep = -np.angle(-rewound_fitted_localsweep)
    ax4.plot(kid['localsweep'].x, phase_localsweep, label='Sweep')
    ax4.plot(kid['localsweep'].x, phase_fitted_localsweep, 'y', lw=3, alpha=0.5, label='fit')

    ## spline interpolation
    try:
        tck = interpolate.splrep(phase_fitted_localsweep, kid['localsweep'].x, s=0)
        x = interpolate.splev(phase_tod, tck, der=0)
        ax4.plot(x, phase_tod, '.', alpha=0.2, label='TOD')
    except ValueError:
        print('KID[{}] TOD interpolation failed..'.format(kid['kidid']))

    tck = interpolate.splrep(kid['localsweep'].x, phase_localsweep, s=0)
    y = interpolate.splev(kid['tod'].frequency, tck, der=0)
    ax4.plot(kid['tod'].frequency, y, 'go', label='carrier f')
    y = interpolate.splev(fr, tck, der=0)
    ax4.plot(fr, y, 'r*', label='resonance f')
    ax4.set_title('KID[{}], Phase vs Freq'.format(kid['kidid']))
    ax4.set_xlabel('Frequency [GHz]')
    ax4.set_ylabel('Phase [rad]')
    ax4.set_xlim(min(kid['localsweep'].x), max(kid['localsweep'].x))
    ax4.legend(loc='best')
    ax4.grid()

    #SaveFits.py::Plot()から抜粋
    fc = kid['tod'].frequency #carrier frequency
    tck = interpolate.splrep(kid['localsweep'].x, phase_localsweep, s=0)
    yfc = interpolate.splev(fc, tck, der=0) # phase of carrier f
    yfr = interpolate.splev(fr, tck, der=0) # phase of resonance f

    tck = interpolate.splrep(phase_fitted_localsweep, kid['localsweep'].x, s=0)
    f = interpolate.splev(yfc, tck, der=0)
    linyfc = 4.0*Qr*(f - fr)/fr

    rewound_tod_amplitude = kid['rewound_tod'].amplitude
    rewound_tod_phase = kid['rewound_tod'].phase

    f = interpolate.splev(rewound_tod_phase, tck, der=0)
    linphase = 4.0*Qr*(f - fr)/fr

    ts = kid['tod_timestamp'][:-100]#データの最後の部分にはよくないデータが含まれる可能性があるので、最後の100個ぐらいを除外する(make_reduced_fitsで処理されている)。
    ts = ts - ts[0]

    fig2 = plt.figure(figsize=(8,10))
    ax1 = plt.subplot(311)
    ax2 = plt.subplot(312)
    ax3 = plt.subplot(313)

    ax1.plot(ts, rewound_tod_amplitude, 'r', label='ampl. (from cache)')
    ax1.plot(ts, linphase, 'b', label='linearized phase (from cache)')
    ax1.plot(ts, rewound_tod_phase, 'b', lw=3, alpha=0.3, label='phase (from cache)')
    ax1.plot(ts, np.ones(len(ts))*linyfc, 'c--', lw=1, label='carrier linphase (from cache)')
    ax1.plot(ts, np.ones(len(ts))*yfc, 'c--', lw=3, alpha=0.3, label='carrier phase (from cache)')
    #ax1.set_ylim(-2, 2)
    ax1.set_title('KID[{}]'.format(kid['kidid']))
    ax1.set_ylabel('Normalized Response')
    ax1.legend(loc='best')
    ax1.grid()

    ampl = kid['amp']
    phase = kid['ph']
    linphase = kid['linph']
    yfc = kid['yfc']
    linyfc = kid['linyfc']

    ax2.plot(ts, ampl, 'r', label='ampl. (deglitch from fits)')
    ax2.plot(ts, linphase, 'b', label='linearized phase (deglitch from fits)')
    ax2.plot(ts, phase, 'b', lw=3, alpha=0.3, label='phase (deglitch from fits)')
    ax2.plot(ts, np.ones(len(ts))*linyfc, 'c--', lw=1, label='carrier linphase (from fits)')
    ax2.plot(ts, np.ones(len(ts))*yfc, 'c--', lw=3, alpha=0.3, label='carrier phase (from fits)')
    ax2.set_xlabel('Time [s]')
    ax2.set_ylabel('Normalized Response')
    ax2.legend(loc='best')
    ax2.grid()

    # rebin = hdu['READOUT'].header['DSAMPLE']
    # dt = 1/hdu['READOUT'].header['FRAMERT'] * rebin
    rebin = 1
    dt = 1/kid['framert']*rebin
    plot_tod_ratio = 0.25
    size = 2**int(np.floor(np.log2(len(ts)*plot_tod_ratio)))
    f_, ampl_ = demerge.power_spectrum_density(ampl[:size], dt, 7, window=None, overwrap_half=True)
    f_, phase_ = demerge.power_spectrum_density(phase[:size], dt, 7, window=None, overwrap_half=True)
    f_, linphase_ = demerge.power_spectrum_density(linphase[:size], dt, 7, window=None, overwrap_half=True)
    ax3.semilogx(f_, np.log10(ampl_)*10.0, 'r', label='ampl. (from fits)')
    ax3.semilogx(f_, np.log10(linphase_)*10.0, 'b', label='linearized phase (from fits)')
    ax3.semilogx(f_, np.log10(phase_)*10.0, 'g', lw=3, alpha=0.3, label='phase (from fits)')
    ax3.set_xlabel('Frequency [Hz]')
    ax3.set_ylabel('PSD [dBk/Hz]')
    ax3.set_ylim(-110.,-50.)
    ax3.legend(loc='best')
    ax3.grid()

    return fig1, fig2

def main() -> None:
    """
    コマンドライン引数
    ------------------
    args[1] string フィットの結果が格納された一時ファイル(*.pkl)の名前
    args[2] string グラフを保存するディレクトリ名
    """
    args = sys.argv
    filename = args[1]
    graph_dir = args[2]
    with open(filename, 'rb') as f:
        kid = pickle.load(f)
    fig1, fig2 = plot(kid)
    if not fig1 is None:
        plt.figure(fig1.number)
        plt.savefig('{}/sweep_kid{:05}.png'.format(graph_dir, kid['kidid']), transparent=False)
        plt.close()
    if not fig2 is None:
        plt.figure(fig2.number)
        plt.savefig('{}/tod_kid{:05}.png'.format(graph_dir, kid['kidid']), transparent=False)
        plt.close()

if __name__=='__main__':
    main()
